%% An automatic differentiation class
% A class with forward mode automatic differentiation.
% Only a limited number of functions have been implemented.  
% Only works for scalars, vectors, and matrices
classdef autodiff
  properties
    val   % value 
    deriv % derivative, deriv(:,:,i) is dval / dx_i
  end % properties
  methods
    %% Constructor
    function dx = autodiff(x)
      % FIXME: modify so that can get second and higher order
      % derivatives
      if isempty(x) 
        dx.val = [];
        dx.deriv = [];
      else
        if (ndims(x)>2) 
          error('too many dimensions');
        end
        dx.val = x;
        dx.deriv = zeros([size(x) numel(x)]);
        for i=1:numel(x)
          [j1 j2] = ind2sub(size(x),i);
          dx.deriv(j1,j2,i) = 1;
        end
        % deriv(:,i) is derivative of x wrt x(i)
      end
    end % function autodiff
    
    function v = value(x)
      v = x.val;
    end    
    function d = diff(x)
      d = x.deriv;
    end
    
    function c=plus(a,b) 
      c = autodiff([]);
      if (isa(a,'autodiff')) 
        if (isa(b,'autodiff'))
          c.val = a.val + b.val;
          c.deriv = a.deriv + b.deriv;
        else
          c.val = a.val + b;
          c.deriv = a.deriv;
        end
      else
        if (isa(b,'autodiff'))
          c.val = a + b.val;
          c.deriv = b.deriv;
        else
          warning('I thought this was impossible');
          c.val = a + b;
          c.deriv = [];
        end
      end
    end % plus()

    function c=uminus(a) 
      c = a;
      c.val = -c.val;
      c.deriv = -c.deriv;
    end

    function c=minus(a,b)
      c = a+(-b);
    end
    
    function c=mtimes(a,b)
      c = autodiff([]);
      if (isa(a,'autodiff')) 
        if (isa(b,'autodiff'))
          c.val = a.val*b.val;
          c.deriv = zeros([size(c.val) size(a.deriv,3)]);
          for i=1:size(a.deriv,3)
            c.deriv(:,:,i) = a.deriv(:,:,i) *b.val + ...
                a.val*b.deriv(:,:,i);
          end
        else
          c.val = a.val * b;
          c.deriv = zeros([size(c.val) size(a.deriv,3)]);
          for i=1:size(a.deriv,3)
            c.deriv(:,:,i) = a.deriv(:,:,i)*b;
          end
        end
      else
        if (isa(b,'autodiff'))
          c.val = a * b.val;
          c.deriv = zeros([size(c.val) size(b.deriv,3)]);
          for i=1:size(b.deriv,3)
            c.deriv(:,:,i) = a*b.deriv(:,:,i);
          end
        else
          warning('I thought this was impossible');
          c.val = a * b;
          c.deriv = [];
        end
      end       
    end % mtimes()
    
    function c=times(a,b)
      c = autodiff([]);
      if (isa(a,'autodiff'))
        if (isa(b,'autodiff'))
          c.val = a.val.*b.val;
          c.deriv = zeros([size(c.val) size(a.deriv,3)]);
          for i=1:size(a.deriv,3)
            c.deriv(:,:,i) = reshape(a.deriv(:,:,i),size(a.val)) .*b.val + ...
                a.val.*reshape(b.deriv(:,:,i),size(b.val));
          end
        else
          c.val = a.val .* b;
          c.deriv = a.deriv;
          for i=1:size(a.deriv,3)
            c.deriv(:,:,i) = reshape(a.deriv(:,:,i),size(a.val)).*b;
          end
        end
      else
        if (isa(b,'autodiff'))
          c.val = a .* b.val;
          c.deriv = b.deriv;
          for i=1:size(b.deriv,3)
            c.deriv(:,:,i) = a.*reshape(b.deriv(:,:,i),size(b.val));
          end
        else
          warning('I thought this was impossible');
          c.val = a .* b;
          c.deriv = [];
        end
      end       
    end % times()

    function c=rdivide(a,b)
      % this could be improved performance wise, but at least it's short
      if isa(b,'autodiff')    
        b.val = 1./b.val;
        b.deriv = -b.deriv*(b.val).^2;
        c = a * b;
      else
        c = a * (1./b); 
      end
    end % rdivide

    function x=mldivide(A,y) % solves Ax = y for x
      x = autodiff([]);
      if isa(A,'autodiff')
        if isa(y,'autodiff')
          x.val = A.val \ y.val;
          x.deriv = zeros([size(x.val),size(A.deriv,3)]);
          % could be more efficient by storing decomp of A ...
          for i=1:size(x.deriv,3)
            x.deriv(:,:,i) = A.val \ (reshape(y.deriv(:,:,i),size(y.val)) - ...
                                      reshape(A.deriv(:,:,i),size(A.val))*x.val);
          end
        else 
          x.val = A.val \ y;
          x.deriv = zeros([size(x.val),size(A.deriv,3)]);
          % could be more efficient by storing decomp of A ...
          for i=1:size(x.deriv,3)
            x.deriv(:,:,i) = A.val \ (-reshape(A.deriv(:,:,i),size(A.val))*x.val);
          end
        end
      else
        if isa(y,'autodiff')
          x.val = A \ y.val;
          x.deriv = zeros([size(x.val),size(y.deriv,3)]);
          % could be more efficient by storing decomp of A ...
          for i=1:size(x.deriv,3)
            x.deriv(:,:,i) = A \ reshape(y.deriv(:,:,i),size(y.val));
          end
        else 
          warning('I thought this was impossible');
          x.val = A \ y;
          x.deriv = [];
        end
      end
    end % mldivide()

    function ix = inv(x) 
      ix = x;
      ix.val = inv(x.val);
      for i=1:size(x.deriv,3)
        ix.deriv(:,:,i) = -(ix.val*(x.deriv(:,:,i)))*ix.val;
      end
    end % inv
    
    function e = eq(a,b)
      % FIXME: should this be a.deriv==b.deriv too?
     if isa(a,'autodiff')
       if isa(b,'autodiff')
         e = a.val==b.val;
       else 
         e = a.val == b;
       end
     else 
       if isa(b,'autodiff')
         e = a==b.val;
       else 
         e = a == b;
       end
     end
    end % eq

    function xt = ctranspose(x);
      xt = autodiff([]);
      xt.val = x.val';
      xt.deriv = zeros([size(xt.val) size(x.deriv,3)]);
      for i=1:size(x.deriv,3)
        xt.deriv(:,:,i) = reshape(x.deriv(:,:,i),size(x.val))';
      end
    end % ctranspose

    function xt = transpose(x);
      xt = autodiff([]);
      xt.val = x.val.';
      xt.deriv = zeros([size(xt.val) size(x.deriv,3)]);
      for i=1:size(x.deriv,3)
        xt.deriv(:,:,i) = reshape(x.deriv(:,:,i),size(x.val)).';
      end
    end % transpose

    function b = subsref(a,s) 
      if strcmp(s.type,'()') 
        b = autodiff([]);
        b.val = a.val(s.subs{:});
        b.deriv = zeros([size(b.val) size(a.deriv,3)]);
        for i=1:size(b.deriv,3)
          tmp = reshape(a.deriv(:,:,i),size(a.val));
          b.deriv(:,:,i) = tmp(s.subs{:});
        end
      elseif strcmp(s.type,'.')
        b = a.(s.subs{:});
      else
        error(s.type);
      end
    end % subsref
    
    function a = subsasgn(a,s,b)
      if strcmp(s.type,'()') 
        if ~isa(a,'autodiff')
          a = autodiff(a);
          a.deriv = zeros(size(a.val),size(b.deriv,3));
        end
        if isa(b,'autodiff')
          a.val(s.subs{:}) = b.val;
          for i=1:size(a.deriv,3)
            tmp = reshape(a.deriv(:,:,i),size(a.val));
            tmp(s.subs{:}) = b.deriv(:,:,i);
            a.deriv(:,:,i) = tmp;
          end
        else
          a.val(s.subs{:}) = b;
          for i=1:size(a.deriv,3)
            tmp = reshape(a.deriv(:,:,i),size(a.val));
            tmp(s.subs{:}) = 0;
            a.deriv(:,:,i) = tmp;
          end
        end
      else
        error(s.type);
      end
    end % subsasgn
      
    function c = horzcat(a,b)
      if ~isa(a,'autodiff') 
        a = autodiff(a);
        a.deriv = zeros(size(a.val),size(b.deriv,3));
      end
      if ~isa(b,'autodiff')
        b = autodiff(b);
        b.deriv = zeros(size(b.val),size(a.deriv,3));
      end
      c = autodiff([]);
      c.val = [a.val b.val];
      c.deriv = zeros([size(c.val) size(a.deriv,3)]);
      for i=1:size(c.deriv,3)
        c.deriv(:,:,i) = [reshape(a.deriv(:,:,i),size(a.val)) ...
                          reshape(b.deriv(:,:,i),size(b.val))];
      end
    end % horzcat
    
    function c = vertcat(a,b)
      if ~isa(a,'autodiff') 
        a = autodiff(a);
        a.deriv = zeros(size(a.val),size(b.deriv,3));
      end
      if ~isa(b,'autodiff')
        b = autodiff(b);
        b.deriv = zeros(size(b.val),size(a.deriv,3));
      end
      c = autodiff([]);
      c.val = [a.val; b.val];
      c.deriv = zeros([size(c.val) size(a.deriv,3)]);
      for i=1:size(c.deriv,3)
        c.deriv(:,:,i) = [reshape(a.deriv(:,:,i),size(a.val)); ...
                          reshape(b.deriv(:,:,i),size(b.val))];
      end
    end % vertcat    

    function s = sum(a, dim) 
      s = autodiff([]);
      if nargin==1
        s.val = sum(a.val);
        if (size(a.val,1)>1)
          s.deriv = reshape(sum(a.deriv,1),1,size(a.val,2),size(a.deriv,3));
        else 
          s.deriv = reshape(sum(a.deriv,2),size(a.val,1),1,size(a.deriv,3));
        end
      else
        s.val = sum(a.val,dim);
        siz = size(a.deriv);
        siz(dim) = 1;
        s.deriv = reshape(sum(a.deriv,dim),siz);
      end
    end

    function y = log(x)
      y = x;
      y.val = log(x.val);
      for i = 1:size(x.deriv,3)
        y.deriv(:,:,i) = x.deriv(:,:,i)./x.val;
      end
    end

    function y = normcdf(x,mu,sigma)
      if (nargin<3)
        sigma = 1;
      end
      if (nargin<2)
        mu = 0;
      end
      if ~isa(x,'autodiff') || isa(mu,'autodiff') || isa(sigma,'autodiff')
        error('normcdf only supports x as an autodiff object, not mu or sigma');
      end
      y = autodiff([]);
      y.val = normcdf(x.val,mu,sigma);
      pdf = normpdf(x.val,mu,sigma);
      for i = 1:size(x.deriv,3)
        y.deriv(:,:,i) = x.deriv(:,:,i).*pdf;
      end
    end
    
  end % methods
end % classdef autodef
